import pytest
import os
from nagausspy import (check_geometry_equivalence, GaussianCom,
                       clasify_geometries)


ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def geometry1():
    fcom = os.path.join(ACTUAL_PATH, "data/conf1_comp.com")
    return GaussianCom(fcom).geometry

@pytest.fixture
def geometry2():
    fcom = os.path.join(ACTUAL_PATH, "data/conf2_comp.com")
    return GaussianCom(fcom).geometry

@pytest.fixture
def geometry3():
    fcom = os.path.join(ACTUAL_PATH, "data/conf3_comp.com")
    return GaussianCom(fcom).geometry

def test_equal(geometry1, geometry2, geometry3):
    assert (check_geometry_equivalence(geometry1, geometry2,
                                       [2, 3, 7], [11, 19]))

    assert not (check_geometry_equivalence(geometry1, geometry3,
                                           [2, 3, 7], [11, 19]))

    assert not (check_geometry_equivalence(geometry1, geometry3,
                                           [2, 3, 7], [11, 19]))

    assert not (check_geometry_equivalence(geometry3, geometry2,
                                           [2, 3, 7], [11, 19]))
def test_network(geometry1, geometry2, geometry3):
    entrada = {
        "geom1": geometry1,
        "geom2": geometry2,
        "geom3": geometry3,
    }
    expected_out = {
        0: ["geom1", "geom2"],
        1: ["geom3"]
    }

    assert expected_out == clasify_geometries(entrada, [2, 3, 7], [11, 19])



import pytest
import os
import numpy as np
from nagausspy import GaussianLog, CommandLine

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def optlog():
    flog = os.path.join(ACTUAL_PATH, "data/output.log")
    return GaussianLog(flog)

@pytest.mark.skip("Gives error")
def test_steps(optlog):
    assert len(optlog.optimization_steps) == 13
    assert optlog.final_geometry.energy == -3472.90032890



import pytest
import os
from nagausspy import open_file, GaussianCom, GaussianLog

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]
@pytest.fixture
def logfile():
    flog = os.path.join(ACTUAL_PATH, "data/output_freq.log")
    return flog

@pytest.fixture
def comfile():
    flog = os.path.join(ACTUAL_PATH, "data/nincs4_square.com")
    return flog

def test_open(comfile, logfile):
    assert isinstance(open_file(comfile), GaussianCom)
    assert isinstance(open_file(logfile), GaussianLog)



import pytest
import os
from nagausspy import GaussianCom

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def freqcom():
    flog = os.path.join(ACTUAL_PATH, "data/nincs4_freq.com")
    return GaussianCom(flog)

def test_empty():
    assert len(GaussianCom().geometry) == 0

def test_len(freqcom):
    assert len(freqcom.geometry) == 13

def test_scf(freqcom):
    assert freqcom.commandline["scf"]["conver"].value == "9"

def test_save(freqcom):
    freqcom.save(os.path.join(ACTUAL_PATH, "data/test.com"))



"""
Results of the NMR calculation
"""
from collections import defaultdict
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from .frequency import _wide_spetra
from .geometry import GaussianGeometry
from typing import Dict, List, Tuple


class NMR(object):
    """
    Object sotring the results of the Magnetic shielding tensor for
    the NMR calculations

    Parameters
    ----------
    openfile : buffer
        Buffer from were the data will be read
    geometry : GaussianGeometry
        Geometry with the same atoms as the ones in the results

    """

    def __init__(self, openfile, geometry: GaussianGeometry):
        self._nucleus = defaultdict(lambda: [])

        # populate the nucleus before turning it into a "normal" dict
        for index, atom in enumerate(geometry):
            self._nucleus[atom.atomname].append(index)
        self._nucleus = dict(self._nucleus)

        self._tensors = np.zeros((len(geometry), 3, 3))
        self._eigenvalues = np.zeros((len(geometry), 3))
        self._isotropic = np.zeros(len(geometry))
        self._anisotropic = np.zeros(len(geometry))

        self._read_data(openfile)

    def _read_data(self, openfile):
        for index in range(len(self._isotropic)):
            # Read the first line
            splitted = next(openfile).strip().split()
            self._isotropic[index] = float(splitted[4])
            self._anisotropic[index] = float(splitted[7])

            # Start reading the tensor the indexes for the matrix are
            # read such that X->0 Y->1 Z->2

            # *X line
            splitted = next(openfile).strip().split()
            self._tensors[index, :, 0] = np.float_(splitted[1::2])

            # *Y line
            splitted = next(openfile).strip().split()
            self._tensors[index, :, 1] = np.float_(splitted[1::2])

            # *Z line
            splitted = next(openfile).strip().split()
            self._tensors[index, :, 2] = np.float_(splitted[1::2])

            # Read the eigenvalues
            splitted = next(openfile).strip().split()
            self._eigenvalues[index, :] = np.float_(splitted[1:])

    @property
    def tensors(self) -> np.ndarray:
        """
        Returns the tensors for all the nucleus. This can be filtered
        using the nucleus property as mask.
        """
        return self._tensors

    @property
    def nucleus(self) -> Dict[str, List[int]]:
        """
        Dictionary with the index of each nucleus, this can be used to
        filter the other properties.
        """
        return self._nucleus

    @property
    def isotropic(self) -> np.ndarray:
        """
        Returns the isotropìc displacements for all the nucleus. This can
        be filtered using the nucleus property as mask.
        """
        return self._isotropic

    @property
    def anisotropic(self) -> np.ndarray:
        """
        Returns the anisotropìc displacements for all the nucleus. This can
        be filtered using the nucleus property as mask.
        """
        return self._anisotropic

    @property
    def eigenvalues(self) -> np.ndarray:
        """
        Returns the tensors' eigenvalues for all the nucleus. This can be
        filtered using the nucleus property as mask.
        """
        return self._eigenvalues

    def generate_spectra(self, intensities="isotropic", npoints=3000,
                         delta_limits=(None, None), funct="gaussian",
                         nucleus=None, ref=0, wide=20) -> Tuple[np.ndarray, np.ndarray]:
        """
        Genereates a wide espectra from the nmr data.

        Parameters
        ----------
        intensities : str (optional)
            Which kind of intensities will be plotted. the options are
            "isotropic" and "anisotropic". Default = "isotropic"
        npoints : int (optional)
            THe number of points to generate. Default=3000
        delta_limits : tuple(int or None, int or None) (optional)
            The limits for the lower and higher displacements. If None
            the limit will be deduced from the values of the data.
            Default (None, None)
        funct : WideningFunction compatible or str
            The function to be used to widen the peaks. If a string,
            the the matching preset function will be used. The posible
            values are:
                - gaussian: Gaussian widening
                - delta : Deltas on the frecuencies
                          (the npoints parameter will be ignored)
            Default = 'gaussian'
        nucleus : str or None (optional)
            The nucleus to representcalculate the espectra. If None
            all will be calculated. Default None
        ref : float (optional)
            Value of the displacement for the reference. Default 0
        wide : float (optional)
            FWHM of the function (not applicable for delta) in ppm. Default 20.

        Returns
        -------
        x : numpy.ndarray
            The values of the frecuencies axis
        y : numpy.ndarray
            The value of the intensity for each frequency.

        """

        mask = []
        delta_limits = list(delta_limits)
        if nucleus is None:
            mask = list(range(len(self.anisotropic)))
        else:
            mask = self.nucleus[nucleus]

        if intensities == "isotropic":
            ppm = self.isotropic
        elif intensities == "anisotropic":
            ppm = self.anisotropic

        ppm = ppm[mask]
        ppm = ref - ppm
        intensities = np.ones_like(ppm)

        if delta_limits[0] is None:
            delta_limits[0] = ppm.min() - 3*wide

        if delta_limits[1] is None:
            delta_limits[1] = ppm.max() + 3*wide

        return _wide_spetra(ppm, intensities, wide, funct, npoints,
                            delta_limits[0], delta_limits[1])

    def represent_spectra(self, ax=None, plot_options=None, **kwargs):
        """
        Represents the nmr spectra

        Parameters
        ----------
        ax : matplotlib.Axes or None
            The axes were the plot will be made, if None a new axis
            will be generate.
        plot_options: dict or None (optional)
            Dict with options for the plot. Default None
        **kwargs : Any keyword argument of the function generate_spectra

        Returns
        -------
        ax : matplotlib.axes
            The axes were the plot was made

        """
        if plot_options is None:
            plot_options = {}

        # Save if the axes were generated of used as input
        generated_ax = False
        if ax is None:
            _, ax = plt.subplots()
            generated_ax = True

        ax.plot(*self.generate_spectra(**kwargs), **plot_options)

        # If the axes were generated by the function add some labels
        if generated_ax:
            ax.set_xlabel(r"$\delta$  (ppm)")
            ax.set_ylabel("Intensity (a.u.)")
        return ax



import pytest
import os
import numpy as np
from nagausspy import GaussianLog, Units

ACTUAL_PATH = os.path.split(os.path.join(os.path.abspath(__file__)))[0]

@pytest.fixture
def frequencylog():
    flog = os.path.join(ACTUAL_PATH, "data/output_freq.log")
    return GaussianLog(flog)


def test_modes(frequencylog):
    assert len(frequencylog.frequencies.frequencies) == 33

def test_low(frequencylog):
    low = np.array([-2.38580000e+00, -2.28630000e+00, -7.89300000e-01,
                    -3.50000000e-03, 4.10000000e-03, 5.80000000e-03,
                    1.72462000e+01, 2.30836000e+01, 2.97682000e+01])
    assert np.allclose(low, frequencylog.frequencies.low_frequencies)

def test_generate_spectra(frequencylog):
    freq, intensity = frequencylog.frequencies.generate_spectra()
    assert len(freq) == len(intensity)

def test_thermo(frequencylog):
    result = ("Sum of electronic and zero-point Energies = -3472.858647 Ha\n"
              "Sum of electronic and thermal Energies = -3472.842107 Ha\n"
              "Sum of electronic and thermal Enthalpies = -3472.841163 Ha\n"
              "Sum of electronic and thermal Free Energies = -3472.910891 Ha\n"
              "Total CV = 8.013458337383415e-05 Ha/K\n"
              "Translational CV = 4.710638580138423e-06 Ha/K\n"
              "Rotational CV = 4.710638580138423e-06 Ha/K\n"
              "Vibrational CV = 7.07133062135573e-05 Ha/K\n")

    assert result == str(frequencylog.frequencies.thermochemistry)

def test_substract(frequencylog):
    result = ("Sum of electronic and zero-point Energies = 0.0 Ha\n"
              "Sum of electronic and thermal Energies = 0.0 Ha\n"
              "Sum of electronic and thermal Enthalpies = 0.0 Ha\n"
              "Sum of electronic and thermal Free Energies = 0.0 Ha\n"
              "Total CV = 0.0 Ha/K\n"
              "Translational CV = 0.0 Ha/K\n"
              "Rotational CV = 0.0 Ha/K\n"
              "Vibrational CV = 0.0 Ha/K\n")
    delta = frequencylog.frequencies.thermochemistry
    delta = delta - delta

    assert str(delta) == result

def test_product(frequencylog):
    result = ("Sum of electronic and zero-point Energies = -3472.858647 Ha\n"
              "Sum of electronic and thermal Energies = -3472.842107 Ha\n"
              "Sum of electronic and thermal Enthalpies = -3472.841163 Ha\n"
              "Sum of electronic and thermal Free Energies = -3472.910891 Ha\n"
              "Total CV = 8.013458337383415e-05 Ha/K\n"
              "Translational CV = 4.710638580138423e-06 Ha/K\n"
              "Rotational CV = 4.710638580138423e-06 Ha/K\n"
              "Vibrational CV = 7.07133062135573e-05 Ha/K\n")

    delta = frequencylog.frequencies.thermochemistry
    delta = 2*delta - delta

    assert str(delta) == result

def test_change_units(frequencylog):
    result = ("Sum of electronic and zero-point Energies = -9086662.855201548 kJmol\n"
              "Sum of electronic and thermal Energies = -9086619.578633482 kJmol\n"
              "Sum of electronic and thermal Enthalpies = -9086617.108676996 kJmol\n"
              "Sum of electronic and thermal Free Energies = -9086799.550547505 kJmol\n"
              "Total CV = 0.20967048076923078 kJmol/K\n"
              "Translational CV = 0.01232528846153846 kJmol/K\n"
              "Rotational CV = 0.01232528846153846 kJmol/K\n"
              "Vibrational CV = 0.18501990384615385 kJmol/K\n")

    thermo = frequencylog.frequencies.thermochemistry

    thermo.units = Units.kJmol

    assert str(thermo) == result



#!/usr/bin/env python2
"""
Implements a functor to configure widening functions and some spectra
widening functions.
"""
import numpy as np

class WideningFunction(object):
    """
    Creates a functor that behaves as a wideding function. This
    function is defined by a center, an intensity and a widening.

    Parameters
    ----------
    func : function
        Function to be used as a template for the widening. This
        function must take as arguments the x coordiante, the center, the
        intensity and the wide (in this particular order) and retun the
        value of the y coordinate in such point.
    center : float
        The coordiante of the center of the function
    intensity : float
        The value of the intensity of the function
    widening : float
        The wide of the function

    """

    def __init__(self, func, center, intensity, widening):
        self.func = func
        self.center = center
        self.intensity = intensity
        self.widening = widening

    def __call__(self, *args):
        """
        Calculates the value on the function in the point.

        Parameters
        ----------
        x : float or numpy.ndarray
            The value(s) were the function must be evaluated

        Returns
        -------
        y : float or numpy.ndarray (the same as the input)
            The values of the function in the input(s) points(s).

        """
        return self.func(args[0], self.center, self.intensity, self.widening)


def gaussian(x, center, intensity, widening):
    """
    Evaluates a gaussian distribution function in a point.

    Parameters
    ----------

    x : float or numpy.ndarray
        The point(s) where the function will be evaluated
    center : float
        The mean value of the distribution
    intensity : float
        The value of the integral of the density
    widening : float
        The standar deviation of the distribution

    Returns
    -------
    data : float or numpy.ndarray (the same as x)
        The function evaluated at the x points.
    """

    return intensity/np.sqrt(2*np.pi*widening**2)*np.exp(-(x-center)**2/(2*widening**2))

def lorentzian(x, center, intensity, widening):
    """
    Evaluates a gaussian distribution function in a point.

    Parameters
    ----------

    x : float or numpy.ndarray
        The point(s) where the function will be evaluated
    center : float
        The mean value of the distribution
    intensity : float
        The value of the integral of the density
    widening : float
        The HWHM

    Returns
    -------
    data : float or numpy.ndarray (the same as x)
        The function evaluated at the x points.
    """

    return (intensity/np.pi)*(widening/((x-center)**2+widening**2))



import re
from collections import defaultdict
import numpy as np
from scipy.spatial.distance import cdist
from typing import List, Optional
from gaddlemaps.components import AtomGro, Residue

class GaussianAtom(object):
    """
    Stores the info of an atom, with a similar interface to AtomGro
    """
    ATOMIC_NUMBERS = {1: "H", 2: "He", 3: "Li", 4: "Be", 5: "B", 6:
                      "C", 7: "N", 8: "O", 9: "F", 10: "Ne", 11: "Na",
                      12: "Mg", 13: "Al", 14: "Si", 15: "P", 16: "S",
                      17: "Cl", 18: "Ar", 19: "K", 20: "Ca", 21: "Sc",
                      22: "Ti", 23: "V", 24: "Cr", 25: "Mn", 26: "Fe",
                      27: "Co", 28: "Ni", 29: "Cu", 30: "Zn", 31:
                      "Ga", 32: "Ge", 33: "As", 34: "Se", 35: "Br",
                      36: "Kr", 37: "Rb", 38: "Sr", 39: "Y", 40: "Zr",
                      41: "Nb", 42: "Mo", 43: "Tc", 44: "Ru", 45:
                      "Rh", 46: "Pd", 47: "Ag", 48: "Cd", 49: "In",
                      50: "Sn", 51: "Sb", 52: "Te", 53: "I", 54: "Xe",
                      55: "Cs", 56: "Ba", 57: "La", 58: "Ce", 59:
                      "Pr", 60: "Nd", 61: "Pm", 62: "Sm", 63: "Eu",
                      64: "Gd", 65: "Tb", 66: "Dy", 67: "Ho", 68:
                      "Er", 69: "Tm", 70: "Yb", 71: "Lu", 72: "Hf",
                      73: "Ta", 74: "W", 75: "Re", 76: "Os", 77: "Ir",
                      78: "Pt", 79: "Au", 80: "Hg", 81: "Tl", 82:
                      "Pb", 83: "Bi", 84: "Po", 85: "At", 86: "Rn",
                      87: "Fr", 88: "Ra", 89: "Ac", 90: "Th", 91:
                      "Pa", 92: "U", 93: "Np", 94: "Pu", 95: "Am", 96:
                      "Cm", 97: "Bk", 98: "Cf", 99: "Es", 100: "Fm",
                      101: "Md", 102: "No", 103: "Lr", 104: "Rf", 105:
                      "Db", 106: "Sg", 107: "Bh", 108: "Hs", 109:
                      "Mt", 110: "Ds", 111: "Rg", 112: "Cn", 113:
                      "Nh", 114: "Fl", 115: "Mc", 116: "Lv", 117:
                      "Ts", 118: "Og"}

    DEFAULT_COLORMAP = defaultdict(lambda: "black",
                                   {"N":"blue", "C":"dark grey", "S":"yellow",
                                    "O":"red", "Ni":"green", "H":"white", "B":
                                    "magenta", "Mg":"green", "Na": "orange"})

    def __init__(self, atomnumber: int, position: List[float]):
        self.atomnumber: int = atomnumber
        self.position: np.ndarray = np.array(position)
        self._color = None

    def __str__(self) -> str:
        out = " {:15s} {:13.6f} {:13.6f} {:13.6f}".format(self.atomname,
                                                          *self.position)
        return out

    def __repr__(self) -> str:
        out = " {:2s} Atom at {:8.5f} {:8.5f} {:8.5f}".format(self.atomname,
                                                              *self.position)
        return out

    @property
    def color(self) -> str:
        """
        The color to represent
        """
        if self._color is None:
            return self.DEFAULT_COLORMAP[self.atomname]
        return self._color

    @color.setter
    def color(self, value: str):
        self._color = value

    @property
    def atomname(self) -> str:
        """
        The atomname corresponding to the input atomnumber
        """
        return self.ATOMIC_NUMBERS[self.atomnumber]

    @property
    def xyz_str(self) -> str:
        """
        Representation of tyhe atom in xyz format
        """
        out = "{:>3s}{:15.5f}{:15.5f}{:15.5f}".format(self.atomname,
                                                      *self.position)
        return out


class GaussianGeometry(object):
    """
    Stores the geonetry of a gaussian geometry
    """

    TRIGGER = re.compile("-----------------------------------------------------"
                         "----------------")

    ATOMIC_NAMES = {value: key for key, value in
                    GaussianAtom.ATOMIC_NUMBERS.items()}

    SKIPTRIGGERS = 2
    NTRIGGERS = 3

    def __init__(self, openfile, charge: int=0, multiplicity: int=1):
        self.atoms: List[GaussianAtom] = []
        self._multiplicity = multiplicity
        self._charge = charge
        self._energy = None
        self._load_info(openfile)

    def __repr__(self) -> str:
        return "Geometry with {} atoms".format(self.natoms)

    def __str__(self) -> str:
        out = "{} {}\n".format(self.charge, self.multiplicity)
        out += "\n".join([str(i) for i in self])
        return out

    def __len__(self) -> int:
        return self.natoms

    def __getitem__(self, index) -> GaussianAtom:
        return self.atoms[index]

    def _load_info(self, openfile):
        triggered = 0
        for line in openfile:
            if self.TRIGGER.search(line):
                triggered += 1

                if triggered >= self.NTRIGGERS:
                    break

                continue
            if triggered < self.SKIPTRIGGERS:
                continue
            self._load_line(line)

    def _load_line(self, line: str):
        fields = line.split()
        self.atoms.append(GaussianAtom(int(fields[1]),
                                       [float(i) for i in fields[3:6]]))

    def xyz_str(self, title: Optional[str]=None) -> str:
        """
        Returns the string corresponding to the xyz representation of
        the geometry.

        Parameters
        ----------
        title : str, optional
             The title of the geometry that will appear in the xyz. If
             None a default one will be used.

        Returns
        -------
        xyz : str
            xyz representation of the geometry.
        """

        if title is None:
            title = "Geometry generated by NaGaussPy"
        out = "{}\n{}\n{}\n".format(self.natoms, title,
                                    "\n".join(a.xyz_str for a in self.atoms))
        return out

    def save_as_xyz(self, filename: str, title: Optional[str]=None):
        """
        Saves the geometry as an xyz in a file.

        Parameters
        ----------
        filename : str
            Filename tot he file where the data will be saved.
        title : str, optional
            The title of the geometry that will appear in the xyz. If
            None a default one will be used.

        """
        with open(filename, "w") as openfile:
            openfile.write(self.xyz_str(title=title))

        # Find the bonds

        dist_mat = cdist(self.coordinates, self.coordinates,
                         metric="sqeuclidean")
        dist_mat = dist_mat < cut**2

        for i in range(len(self)-1):
            for j in range(i+1, len(self)):
                if dist_mat[i][j]:
                    mlab.plot3d(*zip(self[i].position,
                                     self[j].position))

    def align_with(self, atom_index: List[int], position: np.ndarray):
        """
        Aligns the geometry ina way that 3 atoms are orientated as 3
        reference points.

        Parameters
        ----------
        atom_index : list
            The indexes of the 3a toms to use in the alignment
        position : np.array(3,3)
            The positions of the reference points in the alignment.
            Each row correspond to one point.


        """
        from gaddlemaps import calcule_base

        old_position = np.array([self[i].position for i in atom_index])

        sel_base, sel_app = calcule_base(position)
        mol_base, mol_app = calcule_base(old_position)

        delta = sel_app - mol_app
        self.coordinates += delta
        conversion_matrix = np.matmul(np.linalg.inv(sel_base), mol_base)
        self.coordinates = np.matmul(conversion_matrix, self.coordinates.T).T

    @classmethod
    def from_xyz(cls, filename: str):
        """
        Creates a geometry from an xyz datafile

        Parameters
        ----------
        filename : str
            XYZ file

        Returns
        -------
        xyz : GaussianGeometry
            The geometry represented on the xyz file
        """

        geometry = cls.__new__(cls)
        geometry.charge = 0
        geometry.multiplicity = 1
        geometry.atoms = []

        with open(filename) as _file:
            # Skip the number of atoms and title line
            _ = next(_file)
            _ = next(_file)

            for line in _file:
                if not line.strip():
                    continue
                atomname, *coords = line.split()[:4]
                coords = np.array([float(i) for i in coords])
                geometry.atoms.append(
                    GaussianAtom(cls.ATOMIC_NAMES[atomname], coords)
                    )
        return geometry

    @property
    def natoms(self) -> int:
        """
        The number of atoms in the system
        """
        return len(self.atoms)

    @property
    def energy(self) -> Optional[float]:
        """
        SCF energy of the configuration in Hartrees
        """
        return self._energy
    @energy.setter
    def energy(self, value: float):
        self._energy = value

    @property
    def charge(self) -> int:
        """
        Charge of the system
        """
        return self._charge
    @charge.setter
    def charge(self, value: int):
        self._charge = value

    @property
    def multiplicity(self) -> int:
        """
        Multiplicity of the system
        """
        return self._multiplicity
    @multiplicity.setter
    def multiplicity(self, value: int):
        self._multiplicity = value

    @property
    def coordinates(self) -> np.ndarray:
        """
        The coordinates of all the atoms
        """
        return np.array([i.position for i in self.atoms])

    @coordinates.setter
    def coordinates(self, newval: np.ndarray):
        if len(newval) == self.natoms:
            for index, coordinate in enumerate(newval):
                self.atoms[index].position = np.array(coordinate)
        else:
            text = ("The number of atoms do not match the dimension "
                    "of the current data ({})")
            raise ValueError(text.format(self.coordinates.shape))

    @property
    def atom_numbers(self) -> np.ndarray:
        """
        The atomic numbers of all the atoms
        """
        return np.array([i.atomnumber for i in self.atoms])

    @property
    def moleculegro(self) -> Residue:
        """
        The MoleculeGro version of the geometry.
        """
        atoms: List[AtomGro] = []
        index = 1
        for number, coordinates in zip(self.atom_numbers, self.coordinates):
            atoms.append(AtomGro([1, "MOL",
                                  GaussianAtom.ATOMIC_NUMBERS[number],
                                  index]+list(coordinates/10)))
            index += 1
        mol = sum(atoms[1:], atoms[0])
        return mol

    def __add__(self, other):
        if isinstance(other, GaussianGeometry):
            new = self.__new__(self.__class__)
            new.atoms = self.atoms + other.atoms
            new.charge = self.charge + other.charge
            new.multiplicity = self.multiplicity + other.multiplicity - 1
            return new

        return other + self
    
    @staticmethod
    def bond_distance(atom1: GaussianAtom, atom2: GaussianAtom) -> float:
        """Returns the bonding distance between 2 atoms in Angstroms.

        Args:
            atom1 (GaussianAtom): First Atom
            atom2 (GaussianAtom): Second Atom

        Returns:
            float: Distance Between atoms.
        """
        return np.linalg.norm(atom1.position - atom2.position) 
    
    @staticmethod
    def angle(atom1: GaussianAtom, atom2: GaussianAtom, atom3: GaussianAtom) -> float:
        """Return the angle between 3 atoms in degrees.

        Args:
            atom1 (GaussianAtom): First atom
            atom2 (GaussianAtom): Second (vertex) atom
            atom3 (GaussianAtom): Third Atom

        Returns:
            float: Angle formed between the atoms.
        """
        vec1 = atom1.position - atom2.position
        vec1 /= np.linalg.norm(vec1)
        vec2 = atom3.position - atom2.position
        vec2 /= np.linalg.norm(vec2)
        cos_angle = np.dot(vec1, vec2)
        angle = np.arccos(cos_angle) * 180 / np.pi
        return angle
    
    @staticmethod
    def dihedral_angle(atom1: GaussianAtom, atom2: GaussianAtom, atom3: GaussianAtom,
                 atom4: GaussianAtom) -> float:
        """Returns the proper dihedral angle formed between 4 atoms in degrees.

        Args:
            atom1 (GaussianAtom): First atom
            atom2 (GaussianAtom): Second Atom
            atom3 (GaussianAtom): Third Atom
            atom4 (GaussianAtom): Forth atom

        Returns:
            float: Dihedral angle in degrees.
        """
        # see https://en.wikipedia.org/wiki/Dihedral_angle#In_polymer_physics
        vectors = [
            atom2.position - atom1.position,
            atom3.position - atom2.position,
            atom4.position - atom3.position,
        ]
        reference1 = np.cross(vectors[0], vectors[1])
        reference1 /= np.linalg.norm(reference1)
        reference2 = np.cross(vectors[1], vectors[2])
        reference2 /= np.linalg.norm(reference2)
        
        cos_angle = np.dot(reference1, reference2)
        
        sin_angle = (vectors[1]/np.linalg.norm(vectors[1])).dot(np.cross(reference1, reference2))
        
        angle = np.arctan2(sin_angle, cos_angle) * 180 / np.pi
        
        return angle


class ComGeometry(GaussianGeometry):
    """
    Geometry of .com file or from the initial geometry of a log file
    """

    TRIGGER = re.compile(r"^\s*$")
    SKIPTRIGGERS = 0
    NTRIGGERS = 1

    def _load_line(self, line: str):
        fields = line.split()
        self.atoms.append(GaussianAtom(self.ATOMIC_NAMES[fields[0]],
                                       [float(i) for i in fields[1:4]]))

def colors_codes():
    """
    Defines a color code for diferent colors.

    Define some colors in rgb: Red, greem, blue, black, white, yellow, cian,
    magenta, brown, grey, light blue, light red, dark grey.

    Returns:
    --------
    colors : dictionary
        A dictionary with the colors names and its respective codes.
    """
    colors = {}
    colors['red'] = (1, 0, 0)
    colors['green'] = (0, 1, 0)
    colors['blue'] = (0, 0, 1)
    colors['black'] = (0, 0, 0)
    colors['white'] = (1, 1, 1)
    colors['yellow'] = (1, 1, 0)
    colors['cian'] = (0, 1, 1)
    colors['magenta'] = (1, 0, 1)
    colors['brown'] = (0.588, 0.294, 0)
    colors['grey'] = (0.5, 0.5, 0.5)
    colors['light red'] = tuple([i/255. for i in (255, 102, 102)])
    colors['light blue'] = tuple([i/255. for i in (153, 153, 255)])
    colors['dark grey'] = tuple([i/255. for i in (32, 32, 32)])
    colors['orange'] = tuple([i/255. for i in (255, 165, 0)])
    colors['orangered'] = tuple([i/255. for i in (255, 69, 0)])
    return colors



#!/usr/bin/env python3
"""
Implements a class to read the basic info of a gaussian file
"""
import re
import enum
import numpy as np
from .geometry import GaussianGeometry, ComGeometry
from .frequency import GaussianFrequencies
from ._commandline import CommandLine
from .nmr import NMR
from typing import Optional, Tuple, Union, List
import MDAnalysis

class JobTypes(enum.Enum):
    """
    Different jobtypes available to read
    """
    OPTIMIZATION = 0
    FREQUENCIES = 1
    NMR = 2
    POPULATION = 3

class GaussianLog(object):
    """
    Reads the content of a Gaussian .log file
    """

    def __init__(self, filename: str):
        self.TRIGGERS = {
            re.compile("Input orientation:"): self._read_geom,
            re.compile("Standard orientation:"): self._read_geom,
            re.compile(("Full mass-weighted force "
                        "constant matrix:")): self._read_frequencies,
            re.compile("^ #"): self._read_command_lines,
            re.compile("Symbolic Z-matrix:"): self._read_initial,
            re.compile("Magnetic shielding tensor"): self._read_nmr,
            re.compile("Normal termination of Gaussian"): self._toogle_end,
            re.compile("SCF Done:"): self._read_scf_energy,
        }

        self._commandline = None
        self._initial_geometry = None
        self._optimization_steps = []
        self._frequencies = None
        self._nmr = None
        self._normal_termination = False

        self._job_types = set()

        self._filename = filename
        self._parse()

    def __repr__(self):
        return "Gaussian log file '{}".format(self._filename)


    def _parse(self):
        with open(self._filename, "r") as openfile:
            for line in openfile:
                for trigger, method in self.TRIGGERS.items():
                    if trigger.search(line):
                        method(openfile, line)
        # Copy the nergy of the -2 frame into the last one, if it is
        # an optimization as it is aduplicate
        if JobTypes.OPTIMIZATION in self._job_types:
            self.final_geometry.energy = self.optimization_steps[-2].energy

    def _read_nmr(self, openfile, _):
        self._job_types.add(JobTypes.NMR)
        self._nmr = NMR(openfile, self.final_geometry)

    def _toogle_end(self, *_):
        self._normal_termination = True

    def _read_command_lines(self, openfile, line):
        lines = [line.split("#")[1].strip()]

        line = next(openfile)
        while not line.strip().startswith("-"):
            lines.append(line.strip())
            line = next(openfile)
        self._commandline = CommandLine(" ".join(lines))
        # prevent future reading of more lines starting with #
        self.TRIGGERS[re.compile("^ #")] = lambda *args: None

    def _read_geom(self, openfile, _):
        if self._optimization_steps:
            self._job_types.add(JobTypes.OPTIMIZATION)
        self._optimization_steps.append(GaussianGeometry(openfile,
                                                         charge=self.charge,
                                                         multiplicity=self.multiplicity))

    def _read_scf_energy(self, _, line):
        energy = float(line.split("=")[1].split()[0])
        self.final_geometry.energy = energy


    def _read_initial(self, openfile, _):
        line = next(openfile).split()
        charge = int(line[2])
        multi = int(line[5])
        self._initial_geometry = ComGeometry(openfile, charge, multi)

    def _read_frequencies(self, openfile, _):
        molsize = len(self.initial_geometry.coordinates)
        self._frequencies = GaussianFrequencies(openfile, molsize)

    @property
    def initial_geometry(self) -> Optional[ComGeometry]:
        """
        The initial geometry of the simulation
        """
        return self._initial_geometry

    @property
    def optimization_steps(self) -> Tuple[GaussianGeometry, ...]:
        """
        Returns a tuple with the steps of the optimization
        """
        return tuple(self._optimization_steps)

    @property
    def filename(self) -> str:
        """
        The filename and path of the log file
        """
        return self._filename

    @property
    def frequencies(self) -> Optional[GaussianFrequencies]:
        """
        THe frequencies object storing, if any, the frequencies of the
        log file
        """
        return self._frequencies

    @property
    def nmr(self) -> Optional[NMR]:
        """
        The nmr results, if any.
        """
        return self._nmr

    @property
    def final_geometry(self) -> Optional[ComGeometry]:
        """
        The las geometry stored on the log file
        """
        if JobTypes.OPTIMIZATION in self._job_types:
            return self._optimization_steps[-1]
        return self.initial_geometry

    @property
    def commandline(self) -> Optional[CommandLine]:
        """
        The parsed commandline taht was used to run the log file.
        """
        return self._commandline

    @property
    def charge(self) -> int:
        """
        Charge of the system
        """
        return self.initial_geometry.charge

    @charge.setter
    def charge(self, value: int):
        self.initial_geometry.charge = value
        for geom in self.optimization_steps:
            geom.charge = value

    @property
    def multiplicity(self):
        """
        Multiplicity of the system
        """
        return self.initial_geometry.multiplicity

    @multiplicity.setter
    def multiplicity(self, value: int):
        self.initial_geometry.multiplicity = value
        for geom in self.optimization_steps:
            geom.multiplicity = value

    @commandline.setter
    def commandline(self, value: Union[CommandLine,str]):
        if isinstance(value, CommandLine):
            self._commandline = value
        else:
            self._commandline = CommandLine(value)

    @property
    def normal_termination(self) -> bool:
        """
        Wether the termiantion is normal
        """
        return self._normal_termination

    def write_gro_and_xtc(self, out_basename: str, box: Optional[list[float]]=None,
                          resname: str='mlog', find_energy_minima: bool=False,
                          numbered_names: bool=False, bond_scan: Optional[List[int]]=None):
        """
        Writes a .gro and .xtc with the opt steps to open with VMD.

        Parameters
        ----------
        out_basename: str
            The base name of the output .gro and .xtc files.
        box: list of float
            A list with 3 (lx, ly, lz) or 6 (lx, ly, lz, alpha, beta, gamma)
            floats with the simulation box information. If it is None the box is
            set to a cubic one with side lenght of 1 nm.
        resname: str
            The resname to assign to the atoms in the system (optional).
        numbered_names: bool
            If True an index will be set to each atom to allow a better
            differentiation in VMD.
        find_energy_minima: bool
            If True, only the configurations with local minima energy will be
            considered.
        bond_scan: list of int
            A list with 2 index of the atoms to compute the distance to filter
            the optimization steps.

        """

        n_atoms = len(self._optimization_steps[0])
        universe = MDAnalysis.Universe.empty(n_atoms, 1, 1, trajectory=True)
        universe.add_TopologyAttr('names')
        universe.add_TopologyAttr('resnames')
        if not isinstance(resname, str):
            raise IOError(f'resnames must be a string.')
        universe.atoms.residues[0].resname = resname
        # Set box
        if box is None:
            universe.dimensions = [10, 10, 10, 90, 90, 90]
        elif len(box) == 3:
            universe.dimensions = list(box) + [90, 90, 90]
        else:
            universe.dimensions = box

        # Initial Gro 
        for index, (atom_md, atom_log) in enumerate(zip(universe.atoms,
                                                        self._optimization_steps[0].atoms)):
            if numbered_names:
                name = atom_log.atomname + str(index)
            else:
                name = atom_log.atomname
            atom_md.position = atom_log.position
            atom_md.name = name
        universe.atoms.write(out_basename + '.gro')
        # Trajectory
        opt_steps = self.filter_optimization_steps(find_energy_minima, bond_scan)
        with MDAnalysis.Writer(out_basename + '.xtc', n_atoms) as writer:
            for opt in opt_steps:
                universe.atoms.positions = opt.coordinates
                writer.write(universe.atoms)

    def filter_optimization_steps(self, find_energy_minima: bool=False,
                                  bond_scan: Optional[List[int]]=None) -> List[GaussianGeometry]:
        """
        Filters the optimazation steps searching energy minima or bonds.

        Parameters
        ----------
        find_energy_minima: bool
            If True, only the configurations with local minima energy will be
            considered.
        atoms_index: list of int
            A list with 2 index of the atoms to compute the distance.
        """
        if find_energy_minima:
            filter_steps = energy_minima(self.optimization_steps)
        if bond_scan is not None:
            filter_steps = scan_configurations(self.optimization_steps, bond_scan)
        return filter_steps


def energy_minima(optimization_steps: Tuple[GaussianGeometry]) -> List[GaussianGeometry]:
    """
    Filters the steps to keep only the local energy minima

    Parameters
    ----------
    optimization_steps: list of GaussianGeometry
        The steps to filter.
    """
    
    filter_steps = []
    for index in range(len(optimization_steps) - 1):
        old_opt, new_opt = optimization_steps[index:index+2]
        if old_opt.energy is None:
            continue
        if new_opt.energy > old_opt.energy:
            filter_steps.append(old_opt)
    return filter_steps


def scan_configurations(optimization_steps: Tuple[GaussianGeometry],
                        atoms_index: List[int]) -> List[GaussianGeometry]:
    """
    Filters the steps to keep those with different atomic distances.

    Parameters
    ----------
    optimization_steps: list of GaussianGeometry
        The steps to filter.
    atoms_index: list of int
        A list with 2 index of the atoms to compute the distance.
    """
    filter_steps = []
    for index in range(len(optimization_steps) - 1):
        old_opt, new_opt = optimization_steps[index:index+2]
        if old_opt.energy is None:
            continue
        old_bond = np.linalg.norm(old_opt[atoms_index[0]].position -
                                  old_opt[atoms_index[1]].position)
        new_bond = np.linalg.norm(new_opt[atoms_index[0]].position -
                                  new_opt[atoms_index[1]].position)
        if not np.isclose(new_bond, old_bond):
            filter_steps.append(old_opt)
    return filter_steps



from typing import List
class CommandLine(object):
    """
    Stores the info about the command line of ht egaussian log file

    Attributes
    ----------
    keywords : list(Keyword)
        List with all the keyword of the command line.
    """

    def __init__(self, line=""):
        self.keywords: List[Keyword] = [Keyword(key) for key in line.split()]

    def __str__(self):
        return "# " + " ".join(str(i) for i in self.keywords)

    __repr__ = __str__

    def __getitem__(self, keywordname: str) -> 'Keyword':
        for key in self.keywords:
            if key.keyword == keywordname:
                return key
        raise IndexError

    def add_keyword(self, keywordstr: str):
        """
        Adds a new keyword to the command line

        Parameters
        ----------
        keywordstr : string
            String with the definition of the keyword an its modifiers
            as would appear un the .com file.

        """
        keyword = Keyword(keywordstr)
        try:
            _ = self[keyword.keyword]
        except IndexError:
            self.keywords.append(keyword)
            return

        text = "Keyword {} already in the command line"
        text = text.format(keyword.keyword)
        raise RuntimeError(text)

    def remove_keyword(self, keyword: str):
        """
        Removes one keyword from the command line.

        Parameters
        ----------
        keyword : string
            Name of the keyword to remove

        """
        keyword = self[keyword]
        self.keywords.remove(keyword)

class Keyword(object):
    """
    Stores a gaussian keyword and its modifiers
    """

    def __init__(self, keyword_str: str):
        self._string = keyword_str

        self._keyword = None
        self._modifiers = ModifierSet()
        self._format_string()

    def __str__(self):
        out = self.keyword
        if len(self.modifiers) == 1:
            out += "={}".format(list(self.modifiers)[0])
        elif len(self.modifiers) >= 2:
            out += "=({})".format(",".join(str(i) for i in self.modifiers))
        return out

    __repr__ = __str__

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if self.keyword != other.keyword:
            return False
        if self.modifiers != other.modifiers:
            return False
        return True

    def __getitem__(self, modifiername: str) -> 'Modifier':
        for modifier in self.modifiers:
            if modifier.name == modifiername:
                return modifier
        text = ("Modifier {} is not in the keyword").format(modifiername)
        raise IndexError(text)

    def _format_string(self):
        splitted = self._string.split("=")
        self._keyword = splitted[0]

        if len(splitted) >= 2:
            self._format_modifiers("=".join(splitted[1:]))

    def _format_modifiers(self, modifiers):
        # Remove the parenthesis
        modifiers = modifiers.replace("(", "").replace(")", "")

        # split the modifiers by commas and it to the set of keywords
        for modifier in modifiers.split(","):
            self._modifiers.add(Modifier(modifier))

    def add_modifier(self, modifierstr: str):
        """
        Adds a new modfier to the keyword.

        Parameters
        ----------
        modifierstr : string
            String with the modfier as would be written in the .com file.
        """
        self.modifiers.add(Modifier(modifierstr))

    def remove_modifier(self, modifiername: str):
        """
        Removes a modifier from the keyword.

        Parameters
        ----------
        modifiername : string
            The name of the modifier to remove

        """
        modifier = self[modifiername]
        self.modifiers.remove(modifier)

    @property
    def keyword(self) -> str:
        """
        The main part of the keyword
        """
        return self._keyword

    @property
    def modifiers(self) -> 'ModifierSet':
        """
        A set with the modifiers that apply to the main keyword. Any
        change to this set will reflect on the Keyword object.

        """
        return self._modifiers

    __repr__ = __str__

class ModifierSet(set):
    """
    Stores an indexable set
    """

    # def __init__(self, values=()):
    #     super(ModifierSet, self).__init__(values)

    def __getitem__(self, index: int) -> 'Modifier':
        if index >= len(self):
            raise IndexError

        return tuple(self)[index]

class Modifier(object):
    """
    Stores a modifier and its value (if any).
    """

    def __init__(self, modifier_string: str):
        self._name = None
        self._val = None

        self._parse(modifier_string)

    def _parse(self, modifier_string: str):
        splitted = modifier_string.split("=")
        if len(splitted) == 1:
            self._name = splitted[0].strip()

        elif len(splitted) == 2:
            self._name = splitted[0].strip()
            self._val = splitted[1].strip()

        else:
            text = ("Invalid modifier for keyword {}").format(modifier_string)
            raise ValueError(text)

    def __repr__(self):
        out = str(self.name)
        if self.value is not None:
            out += "={}".format(self.value)
        return out

    __str__ = __repr__

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.value != other.value:
            return False
        return True

    @property
    def name(self) -> str:
        """
        The name of the modifier. This item is not modifiable
        """
        return self._name

    @property
    def value(self):
        """
        The value of the modifier (if any). This value can be midifed
        only if it has a non None value.
        """
        return self._val

    @value.setter
    def value(self, value):
        if self._val is not None:
            self._val = value
        else:
            text = ("The modifier {} seems to be a non-valuable one. If you are"
                    " sure a value can be placed in this modifier, remove it "
                    "from the list an add it as a new one.").format(self.name)
            raise ValueError(text)



from ._commandline import CommandLine
from typing import Dict

COMMANDLINES: Dict[str, CommandLine] = {
    "optimization": CommandLine(("opt B3LYP/6-311g(d,p) "
                                 "int=ultrafine scf=conver=9")),
    "frequency": CommandLine(("freq=raman B3LYP/6-311g(d,p) int=ultrafine "
                              "pop=(full,nbo) scf=conver=9")),
    "population": CommandLine(("B3LYP/6-311g(d,p) int=ultrafine prop "
                               "pop=(full,nbo,esp) scf=conver=9")),
    "nmr": CommandLine(("RHF/6-311G(d,p) nmr"))
}



import re
import enum
import numpy as np
import matplotlib.pyplot as plt
from .widening_functions import WideningFunction, gaussian, lorentzian
from typing import Tuple, Optional


class Units(enum.Enum):
    """
    Some conversion units. All of them are based on eV

    The KbT is the thermal energy at room temperature (278.15K)
    """
    eV = 1.
    Ha = 27.211385
    kCalmol = 0.043
    kJmol = 0.0104
    KbT = 0.02569257


class GaussianFrequencies(object):
    """
    Stores the frequencies of a Gaussian raman spectroscopy
    """

    WIDENING = 20
    _TRANSLATIONS = {"Frequencies": "frequencies",
                     "Red. masses": "reduced_mass",
                     "Frc consts": "force_constants",
                     "IR Inten": "ir_intensities",
                     "Raman Activ": "raman_activity",
                     "Depolar (P)": "depolar_p",
                     "Depolar (U)": "depolar_u"}

    def __init__(self, openfile, molsize: int):

        self.TRIGGERS = {
            re.compile("Low frequencies ---"): self._load_low_freq,
            re.compile("Diagonal vibrational polarizability:"): self._load_polarizability,
            re.compile("                     1"): self._load_main,
            re.compile("- Thermochemistry -"): self._load_thermo,

        }

        self._mode_properties = {
            "frequencies": [],
            "reduced_mass": [],
            "force_constants": [],
            "ir_intensities": [],
            "raman_activity": [],
            "depolar_p": [],
            "depolar_u": [],
        }

        self._normal_modes = np.array([[[]]])
        self._low_frequencies = []
        self._polarizability = np.array([0, 0, 0])
        self._molsize = molsize
        self._thermo = None
        self._load_info(openfile)

    def __repr__(self):
        return ("Frequencies object with "
                "{} normal modes").format(len(self.frequencies))

    @property
    def frequencies(self) -> np.ndarray:
        """
        Returns all the frequencies
        """
        return self._mode_properties["frequencies"].copy()

    @property
    def reduced_mass(self) -> np.ndarray:
        """
        Returns the reduced mass of each of the normal modes
        """
        return self._mode_properties["reduced_mass"].copy()

    @property
    def force_constants(self) -> np.ndarray:
        """
        Returns the force constants of each normal mode
        """
        return self._mode_properties["force_constants"].copy()

    @property
    def ir_intensities(self) -> np.ndarray:
        """
        Returns the intensities of every normal node
        """
        return self._mode_properties["ir_intensities"].copy()

    @property
    def raman_activity(self) -> np.ndarray:
        """
        Returns the raman activity
        """
        return self._mode_properties["raman_activity"].copy()

    @property
    def normal_modes(self) -> np.ndarray:
        """
        Returns all the normal modes. First index the nromal mode,
        second the atom and third the component of the eigenvector
        """
        return self._normal_modes.copy()

    @property
    def low_frequencies(self) -> np.ndarray:
        """
        Returns the frequencies at k=0
        """
        return np.array(self._low_frequencies)

    @property
    def polarizability(self) -> np.ndarray:
        """
        Returns the polarizability on the 3 coordinate axes
        """
        return self._polarizability

    @property
    def thermochemistry(self) -> Optional['GaussianThermochemistry']:
        """
        The object storing all the thermochemistry related values
        """
        return self._thermo

    def _load_polarizability(self, line, openfile):
        line = next(openfile)
        self._polarizability = np.array([float(i) for i in line.split()])
        return

    def _load_main(self, line, openfile):
        self._normal_modes = []

        while line.strip():
            # The previous line should be the one with the indexes
            line = next(openfile) # This line has the atom types

            line = next(openfile)
            while len(line.split("--")) == 2:

                key = line.split("--")[0].strip()
                if key in self._TRANSLATIONS:
                    formated = self._TRANSLATIONS[key]

                    vals = [float(i) for i in line.split("--")[1].split()]

                    self._mode_properties[formated] += vals

                    line = next(openfile)

            # Determine if we have 1, 2, or 3  eigenvalues per line
            nset = int((len(line.split()) - 2) / 3)

            set2append = [np.zeros([self._molsize, 3]) for i in range(nset)]
            for molindex in range(self._molsize):
                line = next(openfile)
                for setindex in range(nset):
                    offset = 2+3*setindex
                    newval = np.array([float(i)
                                       for i in line.split()[offset:offset+3]])
                    set2append[setindex][molindex, :] = newval
            line = next(openfile)  # restart the loop
            self._normal_modes += set2append

        for key, val in self._mode_properties.items():
            self._mode_properties[key] = np.array(val)
        return

    def _load_low_freq(self, line, _):
        fields = line.split("---")[1].split()
        self._low_frequencies += [float(i) for i in fields]
        return

    def _load_thermo(self, _, openfile):
        self._thermo = GaussianThermochemistry(openfile)

    def _load_info(self, openfile):
        for line in openfile:
            for trigger, method in self.TRIGGERS.items():
                if trigger.search(line):
                    method(line, openfile)
                    if method == self._load_thermo:
                        return

    def generate_spectra(self, intensities="IR", npoints=3000, maxfreq=None,
                         funct="gaussian") -> Tuple[np.ndarray,np.ndarray]:
        """
        Generates a wided spectra from the readf requencies.

        Parameters
        ----------
        intensities : str (optional)
            Which kind of intensities will be plotted. the options are
            "IR", "ones", ,and "raman". Default = "IR"
        npoints : int (optional)
            THe number of points to generate. Default=300
        maxfreq : float or None (optional)
            The maximum frequencie of the scpectra. If None, a
            value of 1.1 times the highest freqeuncie normal mode
            will be chosen. Default = None
        funct : WideningFunction compatible or str
            The function to be used to widen the peaks. If a string,
            the the matching preset function will be used. The posible
            values are:
                - gaussian: Gaussian widening
                - delta : Deltas on the frecuencies
                          (the npoints parameter will be ignored)
            Default = 'gaussian'

        Returns
        -------
        x : numpy.ndarray
            The values of the frecuencies axis
        y : numpy.ndarray
            The value of the intensity for each frequency.
        """


        widen = self.WIDENING

        if intensities == "IR":
            data = self.ir_intensities
        elif intensities == "raman":
            data = self.raman_activity
        elif intensities == "ones":
            data = np.ones_like(self.ir_intensities)

        return _wide_spetra(self.frequencies, data, widen,
                            funct, npoints, 0, maxfreq)

    def represent_spectra(self, ax=None, plot_options=None, **kwargs):
        """
        Represents the read spectra

        Parameters
        ----------
        ax : matplotlib.Axes or None
            The axes were the plot will be made, if None a new axis
            will be generate.
        plot_options: dict or None (optional)
            Dict with options for the plot. Default None
        **kwargs : Any keyword argument of the function generate_spectra

        Returns
        -------
        ax : matplotlib.axes
            The axes were the plot was made

        """
        if plot_options is None:
            plot_options = {}

        # Save if the axes were generated of used as input
        generated_ax = False
        if ax is None:
            _, ax = plt.subplots()
            generated_ax = True

        ax.plot(*self.generate_spectra(**kwargs), **plot_options)

        # If the axes were generated by the function add some labels
        if generated_ax:
            ax.set_xlabel(r"Frequency (cm$^{-1}$)")
            ax.set_ylabel("Intensity (a.u.)")
        return ax

class GaussianThermochemistry(object):
    """
    Stores the information about the thermochemistry after a frequencie calculation
    """

    MAX_MAGNITUDES = 4

    def __init__(self, openfile):

        self._TRIGGERS = {
            re.compile("Sum of electronic and zero-point Energies"): self._read_zero,
            re.compile("Sum of electronic and thermal Energies"): self._read_thermal,
            re.compile("Sum of electronic and thermal Enthalpies"): self._read_enthalpy,
            re.compile("Sum of electronic and thermal Free Energies"): self._read_free,
        }

        self._electronic_energy = None
        self._zero_point_energy = None
        self._thermal_energy = None
        self._enthalpy = None
        self._free_energy = None
        self._TS = None
        self._cv = None
        self._cv_rotational = None
        self._cv_translational = None
        self._cv_vibrational = None
        self._units = Units.Ha
        self._parse(openfile)
        self._parse_cv(openfile)

    def __repr__(self):
        if self._electronic_energy is None:
            string = ("Sum of electronic and zero-point Energies = {} {}\n"
                      "Sum of electronic and thermal Energies = {} {}\n"
                      "Sum of electronic and thermal Enthalpies = {} {}\n"
                      "Sum of electronic and thermal Free Energies = {} {}\n")
            string = string.format(self.zero_point_energy, self.units.name,
                                   self.thermal_energy, self.units.name,
                                   self.enthalpy, self.units.name,
                                   self.free_energy, self.units.name)
        else:
            string = ("Electronic energy = {} {}\n"
                      "Zero-point Energy = {} {}\n"
                      "Thermal Energy = {} {}\n"
                      "Thermal Enthalpy = {} {}\n"
                      "Thermal Free Energy = {} {}\n")
            string = string.format(self.electronic_energy, self.units.name,
                                   self.zero_point_energy, self.units.name,
                                   self.thermal_energy, self.units.name,
                                   self.enthalpy, self.units.name,
                                   self.free_energy, self.units.name)
            
        string += ("Total CV = {} {}/K\n"
                   "Translational CV = {} {}/K\n"
                   "Rotational CV = {} {}/K\n"
                   "Vibrational CV = {} {}/K\n")
        string = string.format(self.CV, self.units.name,
                               self.CV_translational, self.units.name,
                               self.CV_rotational, self.units.name,
                               self._cv_vibrational, self.units.name)
        return string

    def _parse(self, openfile):
        read_magnitudes = 0
        for line in openfile:
            for trigger, method in self._TRIGGERS.items():
                if trigger.search(line):
                    method(line)
                    read_magnitudes += 1
                if read_magnitudes == self.MAX_MAGNITUDES:
                    self._finish()
                    return

    def _read_zero(self, line):
        self._zero_point_energy = float(line.split("=")[1])
        return
    
    def _parse_cv(self, openfile):
        for line in openfile:
            if line.strip().startswith("E (Thermal)"):
                break
        next(openfile)
        units_factor = Units.kCalmol.value/(1000*Units.Ha.value)
        self._cv = float(next(openfile).split()[2]) * units_factor
        next(openfile)
        self._cv_translational = float(next(openfile).split()[2]) * units_factor
        self._cv_rotational = float(next(openfile).split()[2]) * units_factor
        self._cv_vibrational = float(next(openfile).split()[2]) * units_factor

        

    def _read_thermal(self, line):
        self._thermal_energy = float(line.split("=")[1])
        return

    def _read_enthalpy(self, line):
        self._enthalpy = float(line.split("=")[1])
        return

    def _read_free(self, line):
        self._free_energy = float(line.split("=")[1])

    def _finish(self):
        self._TS = self._enthalpy - self._free_energy

    def set_energy_reference(self, energy, units=Units.Ha):
        """
        Sets the energy reference to a dsired value taking into
        account the units.

        Parameters
        ----------
        energy : float
            Value of the energy reference.
        units : Units, optional
            Units of the energy reference value. Default Units.Ha

        """
        # If the energy reference already exist undo the reference
        if self._electronic_energy is not None:
            old_energy = self._electronic_energy
            self._electronic_energy = None
            self.set_energy_reference(old_energy, units=self.units)

        # transform to the proper units
        energy += units.value / self.units.value

        self._electronic_energy = energy
        self._zero_point_energy -= energy
        self._thermal_energy -= energy
        self._enthalpy -= energy
        self._free_energy -= energy
        self._TS -= energy

    @property
    def units(self):
        """
        Units of the values stored. Changing this value to another
        will change all the values stored
        """
        return self._units

    @units.setter
    def units(self, value):
        if self._electronic_energy is not None:
            self._electronic_energy *= self.units.value/value.value
        self._zero_point_energy *= self.units.value/value.value
        self._thermal_energy *= self.units.value/value.value
        self._enthalpy *= self.units.value/value.value
        self._free_energy *= self.units.value/value.value
        self._TS *= self.units.value/value.value
        
        self._cv *= self.units.value/value.value
        self._cv_translational *= self.units.value/value.value
        self._cv_rotational *= self.units.value/value.value
        self._cv_vibrational *= self.units.value/value.value
        
        self._units = value

    @property
    def electronic_energy(self):
        """
        SCF electronic energy taken from the reference
        """
        return self._electronic_energy

    @property
    def zero_point_energy(self):
        """
        Returns the corrected zero point energy
        """
        return self._zero_point_energy

    @property
    def thermal_energy(self):
        """
        The corrected thermal energy
        """
        return self._thermal_energy

    @property
    def enthalpy(self):
        """
        The corrected enthalpy
        """
        return self._enthalpy

    @property
    def free_energy(self):
        """
        The corrected Free energy
        """
        return self._free_energy

    @property
    def TS(self):
        """
        The product of the temperature and the
        increment of entropy
        """
        return self._TS
    
    @property
    def CV(self):
        """
        Total Heat capacity at constant volume
        """
        return self._cv
    
    @property
    def CV_translational(self):
        """
        Contribution to the Heat capacity at constant volume due to translation
        """
        return self._cv_translational
    
    @property
    def CV_rotational(self):
        """
        Contribution to the Heat capacity at constant volume due to rotations
        """
        return self._cv_rotational
    
    @property
    def CV_vibrational(self):
        """
        Contribution to the Heat capacity at constant volume due to vibrations
        """
        return self._cv_vibrational

    def __add__(self, other):
        if self.units.value != other.units.value:
            text = "Missmatch in the units of the thermochmistry {} and {}"
            text = text.format(self.units.name, other.units.name)
            raise ValueError(text)

        obj = self.__new__(self.__class__)
        obj._units = self.units
        obj._zero_point_energy = (self.zero_point_energy
                                  + other.zero_point_energy)
        obj._thermal_energy = (self.thermal_energy
                               + other.thermal_energy)
        obj._enthalpy = (self.enthalpy + other.enthalpy)
        obj._free_energy = (self.free_energy + other.free_energy)
        obj._TS = self.TS + other.TS
        obj._cv = self.CV + other.CV
        obj._cv_translational = self.CV_translational + other.CV_translational
        obj._cv_rotational = self.CV_rotational + other.CV_rotational
        obj._cv_vibrational = self.CV_vibrational + other.CV_vibrational

        if (self.electronic_energy is None) and (other.electronic_energy is None):
            obj._electronic_energy = None
        elif self.electronic_energy is None:
            obj._electronic_energy = other.electronic_energy
        elif other.electronic_energy is None:
            obj._electronic_energy = self.electronic_energy
        else:
            obj._electronic_energy = (self._electronic_energy
                                      + other._electronic_energy)
        return obj

    def __mul__(self, other):
        obj = self.__new__(self.__class__)
        obj._units = self.units
        obj._zero_point_energy = self.zero_point_energy * other
        obj._thermal_energy = self.thermal_energy * other
        obj._enthalpy = self.enthalpy* other
        obj._free_energy = self.free_energy * other
        obj._TS = self.TS * other
        obj._cv = self.CV * other
        obj._cv_translational = self.CV_translational * other
        obj._cv_rotational = self.CV_rotational * other
        obj._cv_vibrational = self.CV_vibrational * other

        obj._electronic_energy = self.electronic_energy
        if self.electronic_energy is not None:
            obj._electronic_energy *= other
        return obj

    def __truediv__(self, other):
        return self * (1./other)

    __div__ = __truediv__

    def __rmul__(self, other):
        return self*other

    def __sub__(self, other):
        return self + (-1*other)

def _wide_spetra(frequencies, data, widen, funct, npoints, minfreq, maxfreq):
    functions = {
        "gaussian": gaussian,
        "lorentzian": lorentzian
    }

    if maxfreq is None:
        maxfreq = np.max(frequencies * 1.1)

    if isinstance(funct, str):
        if funct == "delta":
            return _generate_delta_spectra(frequencies, data, maxfreq=maxfreq)
        else:
            funct = functions[funct]

    functs = [WideningFunction(funct, freq, intensity, widen)
              for freq, intensity in zip(frequencies, data)]

    # Create the array of x

    xdata = np.linspace(minfreq, maxfreq, npoints)
    ydata = sum(func(xdata) for func in functs)
    return xdata, ydata

def _generate_delta_spectra(frequencies, data, maxfreq):
    data = data[frequencies < maxfreq]
    frequencies = frequencies[frequencies < maxfreq]
    xdata = np.zeros(3*len(data))
    ydata = np.zeros(3*len(data))

    for i, value in enumerate(data):
        xdata[3*i:3*(i+1)] = frequencies[i]*np.ones(3)
        ydata[1+3*i] = value

    return xdata, ydata



"""
nagausspy: Basics tools for Gaussian
"""
from .templates import COMMANDLINES
from .reader import GaussianLog
from ._com import GaussianCom, GaussianConfig
from .geometry import GaussianGeometry
from .frequency import Units
from ._commandline import CommandLine
from .widening_functions import gaussian
from .analysis import check_geometry_equivalence, clasify_geometries


from . import frequency as frequency
from . import nmr as nmr
from . import reader as reader
from . import widening_functions as widening_functions
from . import templates
from . import analysis as analysis

def open_file(filename):
    """
    Opens a compatible gaussian file choosing the correct class by
    looking at the file extension.

    Parameters
    ----------
    filename : str
        The path to the file you want to open

    """

    extensions = {
        "com": GaussianCom,
        "log": GaussianLog
    }
    cut = filename.split(".")
    if len(cut) == 1:
        text = "It was not possible to determine the extension for file {}"
        raise ValueError(text.format(filename))

    extension = cut[-1].lower()

    if extension not in extensions:
        text = "Unknown extension {}"
        raise ValueError(text.format(extension))

    return extensions[extension](filename)

__version__ = "0.1-beta2"

__all__ = ["Units", "GaussianLog", "GaussianCom", "CommandLine", "gaussian",
           "COMMANDLINES", "open_file", "GaussianGeometry",
           "check_geometry_equivalence", "clasify_geometries", "frequency",
           "GaussianConfig", "nmr", "reader", "widening_functions", "templates",
           "analysis"]



from .geometry import ComGeometry
from ._commandline import CommandLine
from typing import List

class GaussianCom(object):
    """
    Object storing the info about a gaussian .com file

    Parameters
    ----------

    filename : str or None (optional)
        The filename of the com file. If None an empty com file will
        be generated. Default None
    """

    def __init__(self, filename=None):
        self._gaussianconfig = GaussianConfig()
        self._commandline = None
        self._title = None
        self._com_geometry = None
        self._extralines = []
        self._filename = filename

        if filename is None:
            self._commandline = CommandLine()
            self._title = "Generated by nagausspy"
            self._com_geometry = ComGeometry(iter([""]))

        else:
            self._load_file()

    def _load_file(self):
        with open(self._filename, "r") as openfile:
            self._load_config_and_command(openfile)

            self._title = next(openfile).strip()

            # Skip an extra blank line
            _ = next(openfile)

            # Load the charge and the multiplicity
            line = next(openfile)
            charge, multiplicity = [int(i) for i in line.split()]

            self._com_geometry = ComGeometry(openfile, charge=charge,
                                             multiplicity=multiplicity)

            self._load_extra_lines(openfile)

    def _load_extra_lines(self, openfile):
        for line in openfile:
            self._extralines.append(line.strip())
        # Remove blank lines at the end
        while (self.extralines) and (not self._extralines[-1]):
            self._extralines.pop()


    def _load_config_and_command(self, openfile):
        ###############
        # LOAD CONFIG #
        ###############
        line = next(openfile).strip()
        while line.startswith("%"):
            line = line[1:]
            key, value = line.split("=")
            self._gaussianconfig[key] = value
            line = next(openfile).strip()

        ####################
        # LOAD COMMANDLINE #
        ####################
        while not line.startswith("#"):
            line = next(openfile).strip()
        if not line.startswith("#"):
            text = "Gaussian Command line not found after the last config line"
            raise RuntimeError(text)
        line = line[1:].strip()
        commandlines = []
        while line:
            commandlines.append(line)
            line = next(openfile).strip()
        self._commandline = CommandLine(" ".join(commandlines))

    def save(self, filename: str):
        """
        Saves the com object into a file

        Parameters
        ----------

        filename : str
            File where the com object will be written

        """

        with open(filename, "w") as openfile:
            openfile.write(str(self.configuration)+"\n")
            openfile.write(str(self.commandline)+"\n")
            openfile.write("\n")
            openfile.write(self.title+"\n")
            openfile.write("\n")
            openfile.write(str(self.geometry)+"\n")
            openfile.write("\n")

            if self.extralines:
                openfile.write("\n".join(self.extralines))
                openfile.write("\n\n")

    @property
    def configuration(self) -> 'GaussianConfig':
        """
        Returns the lines of the cofniguration (the ones that stat with %)
        """
        return self._gaussianconfig

    @configuration.setter
    def configuration(self, value: 'GaussianConfig'):
        self._gaussianconfig = value

    @property
    def commandline(self) -> CommandLine:
        """
        Commandline object of the cm file
        """
        return self._commandline

    @commandline.setter
    def commandline(self, value: CommandLine):
        self._commandline = value

    @property
    def title(self) -> str:
        """
        Title of the gaussian job
        """
        return self._title

    @title.setter
    def title(self, value: str):
        self._title = value

    @property
    def geometry(self) -> ComGeometry:
        """
        The geometry stored on the file
        """
        return self._com_geometry

    @geometry.setter
    def geometry(self, value: ComGeometry):
        self._com_geometry = value

    @property
    def extralines(self) -> List[str]:
        """
        List with the extra configuration lines after the geometry
        """
        return self._extralines

    @extralines.setter
    def extralines(self, value: List[str]):
        self._extralines = value

    @property
    def charge(self) -> int:
        """
        Charge of the geometry defined in the file
        """
        return self.geometry.charge

    @charge.setter
    def charge(self, value: int):
        self.geometry.charge = value

    @property
    def multiplicity(self) -> int:
        """
        Multiplicity of the geometry defined in the file
        """
        return self.geometry.multiplicity

    @multiplicity.setter
    def multiplicity(self, value: int):
        self.geometry.multiplicity = value


class GaussianConfig(dict):
    """
    Object storing the configuration for the gaussian execution
    """

    def __repr__(self):
        options = ["%{}={}".format(key, value) for key, value in self.items()]
        return "\n".join(options)
    __str__ = __repr__



import numpy as np
import networkx

def clasify_geometries(geometries, align_index,
                       reference_indexes, tolerance=1):
    """
    Filters the input geometries to group the identical ones.

    Note
    ----
    All the geometries must correspond to the same system and with the
    atoms in the same order

    Parameters
    ----------
    geometries : dict(label: nagausspy.GaussianGeometry)
        Dictionary with the geometries to clasify and a label to
        identify them. They key of the dictionary is the label and the
        values are the geometries.

    align_index : list(int)
        List with 3 indexes of the atoms that will be used to align
        the different geometries. The atoms should be chosen such that
        they are in the same relative position in the different
        geometries. Moreover they can't be 3 atoms in the same line.

    reference_indexes : list(int)
        List with the indexes of the atoms than will be used to
        identify the structures

    tolerance : float
        Max displacement of each one of the reference atoms in angstroms.

    Returns
    -------
    result : dict(structure: list(label)) - optional
        Dictionary with the diferent structures and which labels faell
        in that category. Default 1.

    """

    connection = np.zeros((len(geometries), len(geometries)))

    labels, geoms = list(zip(*geometries.items()))

    for index1, geom1 in enumerate(geoms):
        for index2, geom2 in enumerate(geoms):
            if index1 >= index2:
                continue
            if check_geometry_equivalence(geom1, geom2, align_index,
                                          reference_indexes,
                                          tolerance=tolerance):
                connection[index1, index2] = 1
                connection[index2, index1] = 1

    graph = networkx.from_numpy_array(connection)
    clusters = networkx.connected_components(graph)
    out = {}
    for index, geoindexes in enumerate(clusters):
        out[index] = []
        for geoindex in geoindexes:
            out[index].append(labels[geoindex])
    return out

def check_geometry_equivalence(geometry1, geometry2, align_index,
                               reference_indexes, tolerance=0.8):
    """
    geometry1, geometry2 : nagausspy.GaussianGeometry
        The 2 geometries to compare.
    align_index : list(int)
        List with 3 indexes of the atoms that will be used to align
        the different geometries. The atoms should be chosen such that
        they are in the same relative position in the different
        geometries. Moreover they can't be 3 atoms in the same line.

    reference_indexes : list(int)
        List with the indexes of the atoms than will be used to
        identify the structures

    tolerance : float
        Max displacement of each one of the reference atoms in angstroms.

    Returns
    -------
    result : bool
        Wether the 2 geometries are equivalent or not
    """

    geometry2.align_with(align_index,
                         geometry1.coordinates[align_index])

    delta = (geometry2.coordinates[reference_indexes]
             - geometry1.coordinates[reference_indexes])

    return np.all(np.linalg.norm(delta, axis=1) < tolerance)
